def chord(root, intervals):
    """Return a list of all MIDI note values which belong to the chord
    specified by root and intervals."""
    root %= 12
    chromatics = [root]
    for interval in intervals:
        chromatics.append(chromatics[-1] + interval)
    all_octaves = [n for n in range(128) if n % 12 in chromatics]
    return all_octaves


def major(root):
    intervals = [0, 4, 3]
    return chord(root,  intervals)


def minor(root):
    intervals = [0, 3, 4]
    return chord(root,  intervals)


def dim(root):
    intervals = [0, 3, 3]
    return chord(root,  intervals)


def aug(root):
    intervals = [0, 5, 2]
    return chord(root, intervals)


def maj7(root):
    intervals = [0, 4, 3, 4]
    return chord(root,  intervals)


def dom7(root):
    intervals = [0, 4, 3, 3]
    return chord(root,  intervals)


def min7(root):
    intervals = [0, 3, 4, 3]
    return chord(root,  intervals)


if __name__ == "__main__":
    c7 = min7(60)
    for i in range(0, len(c7), 4):
        print(c7[i:i+4])



def guido(lyric, gamut=None, seed=None):
    """Generate text-to-pitches method of Guido d'Arezzo using randomness."""
    lyric = lyric.upper()
    vowels = [char for char in lyric if char in "AEIOU"]

    if gamut is None:                     # historical origin of the word gamut
        gamut = [55, 57, 59, 60, 62,         # list indices will be taken mod 5
                 64, 65, 67, 69, 71,
                 72, 74, 76, 77, 79,
                 81]

    note_assignment = {"A": gamut[0::5],
                       "E": gamut[1::5],
                       "I": gamut[2::5],
                       "O": gamut[3::5],
                       "U": gamut[4::5]}

    def weigh(potential_notes, prev_note):
        if prev_note is None:
            return [1 for note in potential_notes]
        weights = [1/max(abs(note - prev_note), 1/2)      # avoid division by 0
                   for note in potential_notes]
        return weights

    notes = []
    prev_note = None
    noise = rng(seed)

    # not sure how to implement next block this way

    for char in vowels:
        potential_notes = note_assignment[char]
        weights = weigh(potential_notes, prev_note)
        new_note = random.choices(potential_notes, weights, k=1)[0]
        notes.append(new_note)
        prev_note = new_note
    return list_reader(notes)



from datetime import datetime
from midiutil import MIDIFile


def make_cube(pitch, time, note_len, vel, t):
    """Freeze callables and return a cube."""
    frozen_pitch = pitch
    if callable(pitch):
        frozen_pitch = pitch()
    frozen_vel = vel
    if callable(vel):
        frozen_vel = vel()

    cube = None
    if frozen_vel and frozen_pitch is not None:           # no cube if vel == 0
        cube = [frozen_pitch, time, note_len, frozen_vel]
    return cube


def make_freezer(note_len=1/16, gate=1, nudge=0):
    note_len *= 4                     # midiutil measures time in quarter notes

    def freezer(pitch, vel, time, s, t):
        ice_tray = []
        cube_time = time + nudge         # push cube off-grid and maintain time
        cube_len = note_len * gate       # adjust cube length and maintain time
        cube = make_cube(pitch, cube_time, cube_len, vel, t)
        time += note_len
        if cube is not None:
            ice_tray.append(cube)
        return ice_tray, time

    return freezer


def freeze_song(song, filename=None, track_names=None, combine_tracks=False):
    if filename is None:
        dt = datetime.now()
        filename = dt.strftime("%Y-%m-%d_%H%M%S")

    output_file = MIDIFile()
    if combine_tracks:
        output_file.addTrackName(0, 0, filename)

    default_freezer = make_freezer(note_len=1/16)       # tracks can share this

    for track_number, track in enumerate(song):
        if combine_tracks:
            track_number = 0           # just dump all the ice buckets together

        track_name = f"Track {track_number}"
        if track_names:
            track_name = track_name[track_number]
        track_channel = 0

        time = 0
        if not combine_tracks:
            output_file.addTrackName(track_number, time, track_name)
        ice_bucket = []

        for section in track:
            s = 0                                              # phrase counter

            for phrase in section:
                if callable(phrase):
                    phrase = phrase(s)
                t = 0                                            # note counter

                for note in phrase:
                    if note is None:
                        note = [None, 0, None]
                    if type(note) is int:
                        pitch = note
                        vel = 80
                        freezer = None
                        note = [pitch, vel, freezer]

                    pitch, vel, freezer = note
                    if freezer is None:
                        freezer = default_freezer
                    ice_tray, time = freezer(pitch, vel, time, s, t)
                    ice_bucket.extend(ice_tray)
                    t += 1
                s += 1

    for cube in ice_bucket:
        pitch, time, note_len, vel = cube
        output_file.addNote(track_number, track_channel,
                            pitch, time, note_len, vel)

    filename = f"{filename}.mid"
    with open(filename, 'wb') as outf:
        output_file.writeFile(outf)
        print(f"Wrote to file {filename}.")


def freeze_track(track, name=None):
    song = [track]
    freeze_song(song, filename=name, track_names=[name])


def freeze_section(section, name=None):
    track = [section]
    freeze_track(track, name)


def freeze_phrase(phrase, name=None):
    section = [phrase]
    freeze_section(section, name)


if __name__ == "__main__":
    names = [name for name in dir() if not name.startswith("_")]
    imports = ["MIDIFile", "datetime"]
    for name in imports:
        names.remove(name)
    print(", ".join(names))



from additives import list_reader, state_machine, sip_water, rng
from math import sin, pi, pow


# undecorated machines and their derivatives


def contour(iter, init=None, smoothing=1, seed=None, truncate=-1):
    """Generate a random contour via midpoint displacement."""

    noise = rng(seed=seed)
    if init is None:
        vals = [noise(), noise()]

    for i in range(iter):
        new_vals = []

        *head, _ = vals
        _, *tail = vals
        pairs = zip(head, tail)

        for a, b in pairs:
            new_vals.append(a)
            midpoint = (a + b)/2
            displacement = (2*noise() - 1) * pow(2, -smoothing*(i + 1))
            new_vals.append(midpoint + displacement)
        new_vals.append(vals[-1])
        vals = new_vals

    vals = vals[:truncate]

    offset = min(vals)
    vals = [val + offset for val in vals]
    multiplier = 1/max(vals)
    vals = [multiplier * val for val in vals]

    return list_reader(vals)


def euclid(k, n):
    """Generate pattern of k 1's equally spaced between (n-k) 0's."""

    if k == n:
        trigs = [1 for _ in range(n)]
    else:
        trigs = []
        old_res, new_res = -k % n, 0

        for _ in range(n):
            trigs.append(int(old_res > new_res))
            old_res, new_res = new_res, (new_res + k) % n

    return list_reader(trigs)


def fractioning(a, b):
    """Generate Schillinger's fractioning of a and b."""

    if a < b:
        raise ValueError("b cannot exceed a")
    trigs = [0 for t in range(a*a)]
    for i in range(a - b + 1):
        for j in range(a):
            trigs[i*a + j*b] = 1
    return list_reader(trigs)


def sweep(start, end, steps):
    jump = (end - start)/steps
    vals = [start + jump*i for i in range(steps)]

    return list_reader(vals)


def ramp(steps): return sweep(start=0, end=1, steps=steps)


def saw(steps): return sweep(start=1, end=0, steps=steps)


def resultant(a, b):
    """Generate Schillinger's resultant of a and b."""

    trigs = [(t % a == 0) or (t % b == 0) for t in range(a*b)]
    return list_reader(trigs)


def sine(steps, offset=0):
    step = 2*pi/steps
    vals = [sin(step*i + offset) for i in range(steps)]
    return list_reader(vals)


# simple decorated machines and their derivatives


@state_machine
def automaton(row_0=None, rule=30, seed=None):
    """Generate an elementary cellular automaton with wraparound."""
    if row_0 is None:
        row_0 = [int(i == 0) for i in range(8)]
    row = row_0

    def unpack(rule):
        """Calculate binary representation, but backwards to work with list
        indexing."""
        results = [rule >> i & 1 for i in range(8)]
        return results

    results = unpack(rule)

    def iteration(row):
        def look_next_door(index, row):
            left = row[index - 1]
            center = row[index]
            right = row[(index + 1) % len(row)]

            return 4*left + 2*center + right

        new_row = []

        for index, _ in enumerate(row):
            bit = results[look_next_door(index, row)]
            new_row.append(bit)

        return new_row

    while True:
        yield row
        row = iteration(row)


@state_machine
def circle_map(x_0, omega, coupling):
    """Generate the standard circle map."""
    tau = 2*pi
    tau_inv = 1/tau

    x = float(x_0) % 1

    while True:
        yield x
        x = x + omega + tau_inv*sin(tau*x)
        x %= 1


@state_machine
def duffing(x_0, y_0, a=2.75, b=0.2):
    """Generate the Duffing map."""
    x, y = x_0, x_0

    while True:
        yield x
        x, y = y, -b*x + a*y - y**3


@state_machine
def gingerbread(x_0, y_0, a=1, b=1):
    """Generate the Gingerbreadman map."""
    x, y = x_0, y_0

    while True:
        yield x
        x, y = 1 - a*y + b*abs(x), x


@state_machine
def henon(x_0, y_0, a=1.4, b=0.3):
    """Generate the Henon map."""
    x, y = x_0, y_0

    while True:
        yield x
        x, y = 1 - a*x**2 + y, b*x


@state_machine
def lfsr(n_0=1):
    """Generate a 16-bit linear feedback shift register."""
    modulus = 1 << 16
    n = n_0 % modulus
    n += int(n == 0)                                    # don't initialize on 0
    while True:
        yield n
        bit = (n ^ (n >> 1) ^ (n >> 3) ^ (n >> 12)) & 1
        n = (n >> 1) | (bit << 15)
        n %= modulus


@state_machine
def logistic(x_0, r=3.56995):
    """Generate the logistic map."""
    x = x_0

    while True:
        yield x
        x = r*x*(1-x)


@state_machine
def pfsr(n=0b10101010, len=8, prob=0.5, seed=None):
    """Return a generator for a probabalistic feedback shift register."""
    mask = (2 << len) - 1
    n &= mask
    noise = rng(seed=seed)

    while True:
        yield n
        bit = n & 1
        n >>= 1
        if prob == 1 or noise() < prob:
            bit = 1 - bit
        n += bit << (len-1)


@state_machine
def tent(x_0, m=1.5):
    """Generate the tent map."""
    x = x_0

    while True:
        yield x
        x = m*min(x, 1-x)


@state_machine
def xshift(n_0):
    """Generate an xor shift pseudorandom number generator."""
    n = n_0
    len = 16
    modulus = 1 << len
    while True:
        yield n
        n ^= n >> 7
        n ^= n << 9
        n ^= n >> 13
        n %= modulus


# machines which input other machines, and their derivatives


@state_machine
def clip(machine, min, max):
    instance = machine()
    while True:
        yield max(min, min(instance(), max))


@state_machine
def is_over(machine, threshold):
    instance = machine()
    while True:
        yield int(instance() > threshold)


def pulse(steps, duty=0.5):
    instance = sweep(1, 0, steps)
    return is_over(instance, duty)


@state_machine
def attenuvert(machine, mult, offset=0):
    instance = machine()
    while True:
        yield mult*instance() + offset


def offset(machine, offset): return attenuvert(machine, mult=1, offset=offset)


@state_machine
def skip(machine, batches):
    instance = machine()
    while True:
        yield instance()
        for _ in range(batches - 1):
            instance()


# machines which wrap multiple other machines


@state_machine
def interleave(*machines):
    """Cycle through the outputs of the input machines."""
    if not machines:
        raise ValueError("interleave requires a nonempty list of machines.")
    length = len(machines)
    instances = [machine() for machine in machines]
    i = 0
    while True:
        yield instances[i]()
        i += 1
        i %= length


@state_machine
def mix(*machines):                             # don't colide names with sum()
    if not machines:
        raise ValueError("mix requires a nonempty list of machines.")
    instances = [machine() for machine in machines]
    while True:
        yield sum(instance() for instance in instances)


if __name__ == "__main__":
    sip_water()
    machine = automaton()
    for _ in range(16):
        print(machine())



from additives import list_reader, sip_water


def plain_hunt(row):
    """Genterate plain hunt method permutations of row."""
    length = len(row)
    l_mod_2 = length % 2
    row_copy = row.copy()

    pairs_count = length // 2
    period = 2*length

    rows = [row_copy]
    for t in range(period - 1):
        t_mod_2 = t % 2
        pairs_adjust = t_mod_2 * (1 - l_mod_2)  # don't adjust if length is odd
        for n in range(pairs_count - pairs_adjust):
            index = 2*n + t_mod_2
            row[index], row[index + 1] = row[index + 1], row[index]
        row_copy = row.copy()
        rows.append(row_copy)

    return list_reader(rows)


if __name__ == "__main__":
    sip_water()
    ph = plain_hunt([3, 2, 1])
    for _ in range(8):
        print(ph())



from math import floor


def p_gen(gen):
    """Wrap a parameterized generator in a function call."""

    def wrapper(*args, **kwargs):
        generator = gen(*args, **kwargs)

        def func(*args):
            return next(generator)
        return(func)
    return wrapper


def float_to_CC(x):
    x = max(0, min(x, 127))
    return floor(x)


def quantizer(notes):
    """Make a quantizer function."""
    notes.sort()
    min_note, max_note = min(notes), max(notes)

    def quantize(x):
        if callable(x):
            val = x()
        else:
            val = x

        val = max(val, min_note)

        *head, _ = notes
        _, *tail = notes
        pairs = zip(head, tail)
        for a, b in pairs:
            if a <= val < b:
                midpoint = (a + b)/2
                val = a + (b - a)*int(val > midpoint)
                return val

        return max_note

    return quantize


if __name__ == "__main__":
    @p_gen
    def spam(*args):
        x = 55
        while True:
            yield x
            x += 1.1

    chord = [60, 64, 67]
    Cmaj = quantizer(chord)

    eggs = spam()

    notes = [Cmaj(eggs) for _ in range(16)]
    print(notes)



def oct_scale(root, intervals):
    """Return a list of all MIDI note values which belong to the chord
    specified by root and intervals."""
    root %= 12
    chromatics = [root]
    for interval in intervals:
        chromatics.append(chromatics[-1] + interval)
    all_octaves = [n for n in range(128) if n % 12 in chromatics]
    return all_octaves


# interval between penultimate scale degree and root is not necessary, but
# included below for legibility


# diatonic modes in the usual order


def major(root):
    intervals = [2, 2, 1, 2, 2, 2, 1]
    return oct_scale(root, intervals)


def ionian(root):
    return major(root)


def dorian(root):
    intervals = [2, 1, 2, 2, 2, 1, 2]
    return oct_scale(root, intervals)


def phrygian(root):
    intervals = [1, 2, 2, 2, 1, 2, 2]
    return oct_scale(root, intervals)


def lydian(root):
    intervals = [2, 2, 2, 1, 2, 2, 1]
    return oct_scale(root, intervals)


def mixolydian(root):
    intervals = [2, 2, 1, 2, 2, 1, 2]
    return oct_scale(root, intervals)


def minor(root):
    intervals = [2, 1, 2, 2, 1, 2, 2]
    return oct_scale(root, intervals)


def aeolian(root):
    return minor(root)


def locrian(root):
    intervals = [1, 2, 2, 1, 2, 2, 2]
    return oct_scale(root, intervals)


# other scales


def blues(root):
    intervals = [3, 2, 1, 1, 3, 2]
    return oct_scale(root, intervals)


def messiaen(root, mode):
    """Generate one of Messiaen's modes of limited transposition."""
    modes = [[2, 2, 2, 2, 2, 2],
             [1, 2, 1, 2, 1, 2, 1, 2],
             [2, 1, 1, 2, 1, 1, 2, 1, 1],
             [1, 1, 3, 1, 1, 1, 3, 1],
             [1, 4, 1, 1, 4, 1],
             [2, 2, 1, 1, 2, 2, 1, 1],
             [1, 1, 1, 2, 1, 1, 1, 1, 2, 1]]
    intervals = modes[mode - 1]
    return oct_scale(root, intervals)


if __name__ == "__main__":
    C = major(60)
    for i in range(0, len(C), 7):
        print(C[i:i+8])



from additives import list_reader, rng, state_machine, sip_water
from math import floor


def emphasize_16ths(center_vel):
    # curiously identical to ordered dithering numerators
    stress_16ths = [7, -1, 3, -5, 5, -3, 1, -7, 6, -2, 2, -6, 4, -4, 0, -8]
    vels = [center_vel + stress for stress in stress_16ths]
    return list_reader(vels)


def ppp(): return(emphasize_16ths(8))                         # 0 <= ppp <=  15


def pp(): return(emphasize_16ths(24))                        # 16 <= pp  <=  31


def p(): return(emphasize_16ths(40))                         # 32 <= p   <=  47


def mp(): return(emphasize_16ths(56))                        # 48 <= mp  <=  63


def mf(): return(emphasize_16ths(72))                        # 64 <= mf  <=  79


def f(): return(emphasize_16ths(88))                         # 80 <= f   <=  95


def ff(): return(emphasize_16ths(104))                       # 96 <= ff  <= 111


def fff(): return(emphasize_16ths(120))                     # 112 <= fff <= 127


@state_machine
def random_vel(min, max, seed=None):
    noise = rng(seed)
    while True:
        yield floor(min + (max - min + 1)*noise())


if __name__ == "__main__":
    sip_water()
    machine = mf()
    for _ in range(16):
        print(machine())
